from __future__ import print_function

import errno
import os
import socket
import signal
import struct
import sys
import operator
import time

from collections import namedtuple, deque
from functools import partial
from optparse import OptionParser

import zmq
from zmq.eventloop import zmqstream
from zmq.eventloop.ioloop import IOLoop
from zmq.utils import z85
from tornado.iostream import IOStream, StreamClosedError
from tornado.concurrent import Future
from tornado.tcpclient import TCPClient

from sckt_common import span, get_keys, fork_upstream_proxy
import sckt_client.servers as servers

class EndpointFormatException(Exception):
    pass

def fatal(io, errcode, message):
    global __sckt_client_status
    print(message, file=sys.stderr)
    __sckt_client_status = errcode
    io.stop()

def domainify(pubkey):
    from base64 import b32encode
    return b32encode(z85.decode(pubkey)).strip('=').lower()

def parse_endpoint(descriptor):
    exc = EndpointFormatException("endpoint format: <host>:[0-9]+")

    parts = descriptor.strip().split(':')
    if len(parts) > 2:
        raise exc

    try:
        int(parts[-1]) # check if it's a valid int
    except ValueError:
        raise exc
    else:
        if len(parts) == 1:
            parts.insert(0, 'localhost')

        return parts

def getaddress(host):
    return 'tcp://{}:443'.format(socket.getaddrinfo(host, 443, socket.AF_INET, socket.SOCK_STREAM)[0][-1][0])

class ZmqProtocolMixin(object):
    HEARTBEAT = '\x01'

    def connect_ipc(self, ipcaddr, ident):
        self.socket = zmq.Context.instance().socket(zmq.DEALER)
        self.socket.identity = ident
        self.socket.connect(ipcaddr)

        self.stream = zmqstream.ZMQStream(self.socket)
        self.stream.on_recv(self.on_recv)

    def on_recv(self, data):
        route, src, msg = span(partial(operator.eq, ''), data)
        action = msg.pop(0)

        self.on_message(route, src, action, msg)

    def send(self, route, msg):
        self.stream.send_multipart([''] + route + [''] + msg)

    def send_data(self, route, data):
        self.stream.send_multipart([''] + route + ['', 'DATA', data])

    def heartbeat(self, route):
        self.stream.send_multipart([''] + route + ['', self.HEARTBEAT])

    @property
    def identity(self):
        return self.socket.identity

class PortConnection(object):
    def __init__(self, io_loop=None):
        self.deque = deque()
        self.socket = None
        self.io_loop = io_loop or IOLoop.instance()

    def dq(self, callback=None):
        if callback:
            self.deque.append(callback)
            self.run_dq()
        return self.deque

    def _write(self, *args, **kwargs):
        self.socket.write(*args, **kwargs)

    def write(self, d):
        self.dq(partial(self._write, d, callback=self.flush_dq))

    def close(self):
        self.dq(self.socket.close)

    def run_dq(self):
        self.io_loop.add_callback(self.flush_dq)

    def flush_dq(self, _=None):
        if len(self.deque) > 0 \
           and self.socket \
           and not self.socket.closed():
            self.deque.popleft()()

    def configure_socket(self, disconnect, stream_peer, f):
        exc = f.exception()
        if exc:
            disconnect()
            return

        self.socket = f.result()
        self.socket.read_until_close(callback=disconnect,
                                     streaming_callback=stream_peer)
        self.run_dq()

class ProxyConnection(ZmqProtocolMixin):
    def __init__(self, ipcaddr, endpoint, io_loop=None):
        # using port for ident for now, 1:1 port mapping
        self.connect_ipc(ipcaddr, str(endpoint[1]))
        self.endpoint = endpoint
        self.io_loop = io_loop or IOLoop.instance()
        self.connmap = {}

    def del_conn(self, src):
        del self.connmap[str(src)]

    def set_conn(self, src, conn):
        self.connmap[str(src)] = conn
        return conn

    def conn(self, src):
        return self.connmap[str(src)]

    def disconnect(self, src, _=None):
        self.del_conn(src)
        self.send(src, ["DISCONNECT"])

    def on_message(self, route, src, action, msg):
        if action == "CONNECT":
            send_src = partial(self.send_data, src)
            disc_src = partial(self.disconnect, src)
            peer = self.set_conn(src, PortConnection())

            c = TCPClient().connect(*self.endpoint)
            self.io_loop.add_future(c, partial(peer.configure_socket, disc_src, send_src))

        elif action == "DISCONNECT":
            try:
                self.conn(src).close()
            except KeyError as e:
                pass

        elif action == "DATA":
            try:
                for d in msg:
                    self.conn(src).write(d)
            except KeyError as e:
                print("DATA", e)
                # not connected and no deferqueue
                pass

class Control(ZmqProtocolMixin):
    HEARTBEAT_INTERVAL = 5

    def __init__(self, serverkey, domain, options, endpoints):
        self.domain = domain
        self.ipcaddr = options.ipc
        self.options = options
        self.endpoints = endpoints
        self.router = [serverkey]

        self.connect_ipc(self.ipcaddr, 'CONTROL')

    def start(self):
        self.send(self.router, ['GETIP'])
        self.heartbeat()

    def announce(self, ip, ports):
        self.send(self.router, ['ANNOUNCE', self.domain] + ports)

    def heartbeat(self):
        IOLoop.instance().call_later(self.HEARTBEAT_INTERVAL,
                                     partial(super(Control, self).heartbeat, self.router))

    def on_message(self, route, src, action, msg):
        global __sckt_client_status

        if action == "IP":
            ip = msg[0]
            print('IPv6 address: {}'.format(ip))
            if self.domain != '-':
                print('Domain name: {}'.format(self.domain))

            self.announce(ip, [ProxyConnection(self.ipcaddr, e).identity for e in self.endpoints])

        elif action == 'FAIL_IP':
            fatal(IOLoop.instance(), 8, 'Failed to announce services. Please try restarting the client!')

        elif action == 'FAIL_DOMAIN':
            fatal(IOLoop.instance(), 16, 'Failed to announce services. Try a different domain!')

        elif action == self.HEARTBEAT:
            self.last_seen = int(time.time())
            self.heartbeat()

class AuthError(Exception):
    pass

class Auth(ZmqProtocolMixin):
    def __init__(self, url, serverkey, pubkey, seckey):
        self.ctx = zmq.Context()
        self.socket = self.ctx.socket(zmq.DEALER)
        self.socket.identity = pubkey
        self.socket.curve_publickey = pubkey
        self.socket.curve_secretkey = seckey
        self.socket.curve_serverkey = serverkey
        self.socket.connect(url)
        self.pubkey = pubkey
        self.seckey = seckey
        self.serverkey = serverkey

        self.stream = zmqstream.ZMQStream(self.socket)
        self.stream.on_recv(self.on_recv)

    def destroy(self):
        self.ctx.destroy()

    def ask_credentials(self):
        from getpass import getpass
        print('Please enter your sckt.co credentials:')
        return raw_input('Username: '), getpass()

    def box(self, user, pwd):
        from pysodium import crypto_box, crypto_box_NONCEBYTES, randombytes
        nonce = randombytes(crypto_box_NONCEBYTES)
        return nonce, crypto_box("{}:{}".format(user, pwd), nonce,
                                 z85.decode(self.serverkey), z85.decode(self.seckey))

    def start(self):
        self.stream.send_multipart(['', 'AUTH'] + list(self.box(*self.ask_credentials())))
        self.future = Future()
        return self.future

    def on_recv(self, msg):
        err = None
        if len(msg) > 1:
            err = AuthError('Invalid answer received. Panic')

        if msg[0] != 'SUCCESS':
            err = AuthError('Authentication error')

        if err:
            self.future.set_exception(err)
        else:
            self.future.set_result(True)

        self.stream.close()

def start_auth(cb, *args):
    io = IOLoop.instance()
    def timeout():
        fatal(io, 127, AuthError("Authenticaton timeout"))
        return

    auth = Auth(*args)
    f = auth.start()
    t = io.call_later(5, timeout)
    def result(f):
        auth.destroy()

        exc = f.exception()
        if exc:
            fatal(io, 2, exc.message)
            return

        io.remove_timeout(t)
        io.add_callback(cb)

    io.add_future(f, result)

def start_control(server_url, serverkey, pubkey, seckey, options, endpoints):
    fork_upstream_proxy(options.ipc, server_url, serverkey, (pubkey, seckey))

    domain = options.domain or domainify(pubkey) if not options.nodomain else '-'
    # we map the endpoints here to fail early
    Control(serverkey,
            domain,
            options, map(parse_endpoint, endpoints)).start()

def sigint_handler(sig, frame):
    io = IOLoop.instance()
    io.add_callback_from_signal(
        partial(fatal, io, 0, 'Exiting...'))

def __main(args=None):
    global __sckt_client_status
    # parse the given parameters, or sys.argv smartly
    (options, endpoints) = parser.parse_args(args)
    if len(endpoints) < 1:
        print("You need to specify at least one endpoint!\n", file=sys.stderr)
        parser.print_help()
        __sckt_client_status = 1
        return

    (new_key, pubkey, seckey) = get_keys(options.keyfile)
    control = partial(start_control,
                      getaddress(servers.router_server['address']), servers.router_server['pubkey'],
                      pubkey, seckey, options, endpoints)

    if options.login or new_key:
        start_auth(control,
                    getaddress(servers.auth_server['address']), servers.auth_server['pubkey'],
                    pubkey, seckey)
    else:
        control()

    signal.signal(signal.SIGINT, sigint_handler)
    IOLoop.instance().start()

def main(*args, **kwargs):
    global __sckt_client_status
    __sckt_client_status = 0
    __main(*args, **kwargs)
    sys.exit(__sckt_client_status)

parser = OptionParser(
"usage: %prog [options] <endpoint1> <endpoint...>",

version="%prog 0.1",

description="""
The endpoint string format is <host>:<port>, where
the first part is the IP protocol version,
the second part is the host to connect to,
the third part is the port number on the host to proxy.
""",

epilog="""
Multiple endpoints may be specified, and all of them will be
available on the generated IPv6 address.

The program will output the IPv6 address that it negotiates
with the server on standard output.
""")

parser.add_option('-l', '--login', dest="login", action="store_true",
                  default=False,
                  help="force interactive login on startup")

parser.add_option('-n', '--no-domain', dest="nodomain", action="store_true",
                  default=False,
                  help="don't publish domain name; overrides --domain")

parser.add_option('-d', '--domain', dest="domain", action="store",
                  default="",
                  help="domain name: <domain>.username.sckt.co (default: <publickey>.username.sckt.co>)")

parser.add_option('-i', '--ipc', dest="ipc", action="store",
                  default="ipc:///tmp/sckt.agent",
                  help="local ipc socket")

parser.add_option('-k', '--keyfile', dest="keyfile", action="store",
                  default=os.path.join(os.environ['HOME'],".sckt/keys"),
                  help="local ipc socket")
